"""RPALite is an open source RPA (Robotic Process Automation) library. You can use RPALite via Python or [Robot Framework](https://robotframework.org/) to process various automation tasks.
@author: RPALite contributors
"""
__version__ = "0.0.9"

from .rpalite import RPALite
from .image_handler import ImageHandler
